from simplejson import JSONEncoder

RED = '\033[31m'
RESET = '\033[0m'

def print_error(msg):
    print(f"{RED}ERROR: {msg}{RESET}")
    pass

class ApiError(Exception):
    status_code = 200
    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['status'] = 'error'
        rv['message'] = self.message
        return rv


class StrictEncoder(JSONEncoder):
    def __init__(self, *args, **kwargs):
        kwargs["allow_nan"] = False
        kwargs["ignore_nan"] = True
        super().__init__(*args, **kwargs)


import os
from timeit import default_timer as timer
from types import SimpleNamespace
from typing import Tuple
import json

import allel
import numpy as np
import pandas as pd
import zarr
from sklearn.decomposition import PCA
from sklearn.preprocessing import RobustScaler
import umap

from divbrowse import log
from divbrowse.lib.utils import ApiError


def calculate_mean(slice_of_variant_calls: np.ndarray) -> np.ndarray:
    """Calculate the mean for each variant of a variant matrix array holding the number of alternate alleles

    Note:
        Missing variant calls are excluded from the mean calculation

    Args:
        slice_of_variant_calls (numpy.ndarray): Numpy array representing a variant matrix holding the number of alternate allele calls

    Returns:
        numpy.ndarray: Numpy array holding the means per variant
    """

    slice_of_variant_calls_missing_values_to_nan = np.where(slice_of_variant_calls == -1, np.nan, slice_of_variant_calls)
    return np.nanmean(slice_of_variant_calls_missing_values_to_nan, axis=0) #, keepdims=True


def impute_with_mean(slice_of_variant_calls: np.ndarray) -> np.ndarray:
    """variant matrix array for that missing values should be imputed (replaced) with the mean for the variant

    Args:
        slice_of_variant_calls (numpy.ndarray): Numpy array representing a variant matrix holding the number of alternate allele calls

    Returns:
        numpy.ndarray: Imputed version of the input variant matrix array
    """

    imputed = np.copy(slice_of_variant_calls).astype(np.float32)
    means = calculate_mean(slice_of_variant_calls)
    indices_missing = np.where(slice_of_variant_calls == -1)
    imputed[indices_missing] = np.take(means, indices_missing[1])
    imputed = np.nan_to_num(imputed)
    return imputed


def calc_pca_for_slice_of_variant_calls(slice_of_variant_calls, samples_selected):
    """Calculate a PCA for a variant matrix array

    Args:
        slice_of_variant_calls (numpy.ndarray): Numpy array representing a variant matrix holding the number of alternate allele calls

    Returns:
        numpy.ndarray: PCA result aligned with the sample IDs in the first column
    """

    sample_ids = np.array(samples_selected).reshape((-1, 1)).copy()
    calls_imputed = impute_with_mean(slice_of_variant_calls)
    scaler = RobustScaler()
    calls_imputed_scaled = np.nan_to_num(scaler.fit_transform(calls_imputed))
    start = timer()

    n_components = 10
    if calls_imputed_scaled.shape[1] < n_components:
        n_components = calls_imputed_scaled.shape[1]

    try:
        pca_model = PCA(n_components=n_components, whiten=False, svd_solver='randomized', iterated_power=6).fit(calls_imputed_scaled)
        pca_result = pca_model.transform(calls_imputed_scaled)
        log.debug("==== PCA calculation time: %f", timer() - start)
        pca_result_combined = np.concatenate((sample_ids, pca_result), axis=1)
        return pca_result_combined, pca_model.explained_variance_ratio_

    except ValueError:
        return False




def calc_umap_for_slice_of_variant_calls(slice_of_variant_calls, samples_selected, n_neighbors=15):
    """Calculate UMAP for a variant matrix array

    Args:
        slice_of_variant_calls (numpy.ndarray): Numpy array representing a variant matrix holding the number of alternate allele calls

    Returns:
        numpy.ndarray: PCA result aligned with the sample IDs in the first column
    """

    sample_ids = np.array(samples_selected).reshape((-1, 1)).copy()
    calls_imputed = impute_with_mean(slice_of_variant_calls)

    #start = timer()
    #pca_model = PCA(n_components=2, whiten=False, svd_solver='randomized', iterated_power=6).fit(calls_imputed_scaled)
    #pca_result = pca_model.transform(calls_imputed_scaled)
    #log.debug("==== PCA calculation time: %f", timer() - start)
    #pca_result_combined = np.concatenate((sample_ids, pca_result), axis=1)
    #return pca_result_combined

    start = timer()
    umap_result = umap.UMAP(n_components = 2, n_neighbors=n_neighbors, metric='euclidean', random_state=42).fit_transform(calls_imputed) # , random_state=42, densmap=True  , min_dist=0.5   , dens_lambda=5
    log.debug("==== UMAP calculation time: %f", timer() - start)
    umap_result_combined = np.concatenate((sample_ids, umap_result), axis=1)
    return umap_result_combined




class GenotypeData:
    """Class for managing all genotype data related data structures and methods"""

    def __init__(self, config):

        log.debug("GenotypeData::__init__()")

        self.config = config

        self.datadir = config['datadir']
        path_zarr_variants = self.datadir + config['variants']['zarr_dir']

        if not os.path.exists(self.datadir):
            exit('ERROR: the configured datadir does not exist or is not accessible')

        if not os.path.exists(path_zarr_variants):
            exit('ERROR: the configured path for the Zarr-archive of variants does not exist or is not accessible')

        log.debug("GenotypeData::zarr.open_group()")
        self.callset = zarr.open_group(path_zarr_variants, mode='r')
        log.debug(self.callset.tree(expand=True))

        self.available = {
            'snpeff': False,
            'sample_id_mapping': False
        }

        self._load_data()
        self._setup_sample_id_mapping()
        self._create_chrom_indices()
        self._create_list_of_chromosomes()
    

    def _load_data(self):

        self.available_calldata = list(self.callset['calldata'].array_keys())
        self.available_variants_metadata = list(self.callset['variants'].array_keys())

        # set some shortcut variables
        self.reference_allele = self.callset['variants/REF']
        self.alternate_alleles = self.callset['variants/ALT']
        self.variants_qual = self.callset['variants/QUAL']
        self.calldata = self.callset['calldata/GT']

        # convert chrom, pos and samples from zarr.core.Array to numpy ndarray
        self.chrom = self.callset['variants/CHROM'][:]
        self.pos = self.callset['variants/POS'][:]
        self.samples = self.callset['samples'][:]

        # infer the ploidy from the GT field
        self.ploidy = int(self.callset['calldata']['GT'].ndim) - 1

        # check if SnpEff annotation is included
        if 'ANN' in self.available_variants_metadata:
            self.available['snpeff'] = True

        # Derive distinct chromosome ID's from variant matrix
        self.list_chrom = pd.unique(self.chrom).tolist()

        # Create index for samples
        self.idx_samples = allel.UniqueIndex(self.samples)

        # Create combined index for variant positions
        self.idx_pos = allel.ChromPosIndex(self.chrom, self.pos)

        self.samples_dict = dict(zip(self.samples.tolist(), list(range(0, self.samples.shape[0]))))


    def get_vcf_header(self):
        vcf_header_lines = None
        self._vcf_header_lines_file = self.datadir + '____vcf_export_header_lines____.vcf'
        if os.path.exists(self._vcf_header_lines_file):
            with open(self._vcf_header_lines_file, 'r') as f:
                vcf_header_lines = f.read().splitlines()

        return vcf_header_lines


    def _setup_sample_id_mapping(self):

        if self.config['variants']['sample_id_mapping_filename']:
            self._path_sample_id_mapping = self.datadir + self.config['variants']['sample_id_mapping_filename']
            if os.path.exists(self._path_sample_id_mapping):
                self.available['sample_id_mapping'] = True
            else:
                exit('ERROR: the configured path for the Sample-ID-mapping does not exist or is not accessible')
        
        if self.available['sample_id_mapping']:
            df_sample_id_mapping = pd.read_csv(self._path_sample_id_mapping, header=None)
            self.map_input_sample_ids_to_vcf_sample_ids_dict = dict(zip(df_sample_id_mapping[0], df_sample_id_mapping[1]))
            self.map_vcf_sample_ids_to_input_sample_ids_dict = dict(zip(df_sample_id_mapping[1], df_sample_id_mapping[0]))



    def _create_chrom_indices(self):

        chrom_indices = {}
        for _chr in self.list_chrom:
            cached_index_path = self.datadir + '____pandas_index_chromosome_' + str(_chr) + '_.hdf5'
            try:
                chrom_indices[_chr] = pd.read_hdf(cached_index_path, key='s')
                log.debug("++++ Loaded Pandas index for chromosome %s", str(_chr))

            except FileNotFoundError:
                log.debug("++++ Creating Pandas index for chromosome %s", str(_chr))
                chrom_range = self.idx_pos.locate_key(_chr)
                chrom_indices[_chr] = pd.Series(data=np.arange(chrom_range.start, chrom_range.stop), index=self.pos[chrom_range])
                chrom_indices[_chr].to_hdf(cached_index_path, key='s', mode='w', complevel=5, complib='blosc:zstd')
        
        self.chrom_indices = chrom_indices



    def _create_list_of_chromosomes(self):

        chromosome_labels = {str(key): str(value) for key, value in self.config['chromosome_labels'].items()}
        centromeres_positions = {str(key): str(value) for key, value in self.config['centromeres_positions'].items()}

        path_chromosome_tmp_data = self.datadir+'____list_of_chromosomes____.json'

        try:
            with open(path_chromosome_tmp_data) as f:
                log.debug("++++ Load chromosome metadata from json on disk")
                self.list_of_chromosomes = json.loads(f.read())

        except FileNotFoundError:
            log.debug("++++ Need to create chromosome metadata on thy fly")
            self.list_of_chromosomes = []
            for _chr in self.list_chrom:
                _region = self.idx_pos.locate_range(_chr)
                self.list_of_chromosomes.append({
                    'id': _chr,
                    'label': chromosome_labels[str(_chr)],
                    'centromere_position': int(centromeres_positions[str(_chr)]),
                    'start': int(self.pos[ _region.start ]),
                    'end': int(self.pos[ _region.stop - 1 ]),
                    'number_of_variants': int(self.pos[_region].size)
                })
            with open(path_chromosome_tmp_data, 'w') as outfile:
                json.dump(self.list_of_chromosomes, outfile)


    def sample_ids_to_mask(self, sample_ids: list) -> np.ndarray:
        """Creates a boolean mask based on the input sample IDs that could be found in the samples array of the Zarr storage

        Args:
            sample_ids (list): List with sample IDs

        Returns:
            numpy.ndarray: Boolean mask, True for found sample IDs
        """

        samples_mask = np.zeros(self.samples.shape[0], dtype=bool)
        for sample_id in sample_ids:
            pos = self.samples_dict[sample_id]
            samples_mask[pos] = True
        return samples_mask


    def map_input_sample_ids_to_vcf_sample_ids(self, sample_ids: list) -> list:
        """Map input sample IDs to VCF sample IDs according to the configured mapping table

        Args:
            sample_ids (list): List with sample IDs

        Returns:
            list: List of mapped sample IDs
        """

        mapped_sample_ids = []
        unmapable_sample_ids = []

        if self.available['sample_id_mapping']:
            start = timer()
            for input_sample_id in sample_ids:
                if input_sample_id in self.map_input_sample_ids_to_vcf_sample_ids_dict:
                    mapped_sample_ids.append( self.map_input_sample_ids_to_vcf_sample_ids_dict[input_sample_id] )
                else:
                    unmapable_sample_ids.append(input_sample_id)
            log.debug("==== map_input_sample_ids_to_vcf_sample_ids() calculation time: %f", timer() - start)
        else:
            #mapped_sample_ids = sample_ids
            for input_sample_id in sample_ids:
                if input_sample_id in self.samples_dict:
                    mapped_sample_ids.append(input_sample_id)
                else:
                    unmapable_sample_ids.append(input_sample_id)

        return mapped_sample_ids, unmapable_sample_ids


    def map_vcf_sample_ids_to_input_sample_ids(self, sample_ids: list) -> list:
        """Map VCF sample IDs to input sample IDs according to the configured mapping table

        Args:
            sample_ids (list): List with sample IDs

        Returns:
            list: List of mapped sample IDs
        """
        
        mapped_sample_ids = []
        unmapable_sample_ids = []

        if self.available['sample_id_mapping']:
            start = timer()
            for input_sample_id in sample_ids:
                if input_sample_id in self.map_vcf_sample_ids_to_input_sample_ids_dict:
                    mapped_sample_ids.append(self.map_vcf_sample_ids_to_input_sample_ids_dict[input_sample_id])
                else:
                    unmapable_sample_ids.append(input_sample_id)
            log.debug("==== map_vcf_sample_ids_to_input_sample_ids() calculation time: %f", timer() - start)
        else:
            mapped_sample_ids = sample_ids

        return mapped_sample_ids, unmapable_sample_ids


    def get_samples_mask(self, sample_ids):
        """Returns a tupel consisting of a boolean mask for found sample Ids and a list of mapped sample IDs

        Args:
            sample_ids (list): List with sample IDs

        Returns:
            numpy.ndarray: Boolean mask, True for found sample IDs
            list: mapped sample IDs
        """

        try:
            samples_mask = self.sample_ids_to_mask(sample_ids)
            samples_mapped, unmapable_sample_ids = self.map_vcf_sample_ids_to_input_sample_ids(self.samples[samples_mask].tolist())
        except KeyError:
            raise ApiError('The following sample-IDs could not be resolved: '+', '.join(unmapable_sample_ids))

        return samples_mask, samples_mapped


    def get_posidx_by_genome_coordinate(self, chrom, pos) -> Tuple[int, str]:
        """Returns array coordinates for given physical position on a given chromosome

        Args:
            chrom (str): ID of the chromosome
            pos (int): Physical position on the chromosome

        Returns:
            lookup (int) Array coordinate of the found physical position on the chromosome
            lookup_type (str): Type of the lookup, could be either 'direct_lookup' or 'nearest_lookup'
        """
        
        pd_series = self.chrom_indices[chrom]
        try:
            lookup = pd_series.at[pos]
            return lookup, 'direct_lookup'
        except KeyError:
            # do fuzzy search via nearest neighbor search
            nearest = pd_series.index.get_indexer([pos], method='nearest')[0]
            lookup = pd_series.iloc[nearest]
            return lookup, 'nearest_lookup'


    def count_alternate_alleles(self, sliced_variant_calls):
        """Returns a tupel consisting of a boolean mask for found sample Ids and a list of mapped sample IDs

        Args:
            sliced_variant_calls (numpy.ndarray): variant matrix array holding the allele calls (0/0  0/1  1/1)

        Returns:
            numpy.ndarray: variant matrix array holding the number of alternate allele calls
        """

        # monoploid / haploid
        if sliced_variant_calls.ndim == 2:
            numbers_of_alternate_alleles = allel.HaplotypeArray(sliced_variant_calls).T

        # diploid
        if sliced_variant_calls.ndim == 3:
            # Transform each genotype call into the number of non-reference alleles and then transpose it via .T to row-major order
            numbers_of_alternate_alleles = allel.GenotypeArray(sliced_variant_calls).to_n_alt(fill=-1).T

        return numbers_of_alternate_alleles


    def count_variants_in_window(self, chrom, startpos, endpos) -> int:
        """Counts number of variants in a genomic region

        Args:
            chrom (str): The chromosome of the genomic region.
            startpos (int): The first position of the genommic region.
            endpos (int): The last position of the genommic region.

        Returns:
            int: Number of variants in the genomic region

        """

        location_start, lookup_type_start = self.get_posidx_by_genome_coordinate(chrom, startpos)
        location_end, lookup_type_end = self.get_posidx_by_genome_coordinate(chrom, endpos)
        location_end = location_end + 1
        positions = self.pos[location_start:location_end]

        count = 0
        for i in range(positions.shape[0]):
            if positions[i] >= startpos and positions[i] <= endpos:
                count = count + 1
        
        return count


    def calculate_minor_allele_freq(self, numbers_of_alternate_alleles):
        """Calculates minor allele frequency

        Args:
            numbers_of_alternate_alleles (numpy.ndarray): Numpy array representing a variant matrix holding the number of alternate allele calls

        Returns:
            numpy.ndarray: Numpy array (1d) holding the calculated minor allele frequencies per each variant
        """

        if self.ploidy == 1:
            means = calculate_mean(numbers_of_alternate_alleles)
            maf = np.where(means < 0.5, means, 1 - means)
            return np.nan_to_num(maf, nan=-1).tolist()

        if self.ploidy == 2:
            means_halfed = calculate_mean(numbers_of_alternate_alleles) / 2
            maf = np.where(means_halfed < 0.5, means_halfed, 1 - means_halfed)
            return np.nan_to_num(maf, nan=-1).tolist()


    def calc_variants_summary_stats(self, numbers_of_alternate_alleles):    
        result = {}

        result['maf'] = self.calculate_minor_allele_freq(numbers_of_alternate_alleles)

        df_numbers_of_alternate_alleles = pd.DataFrame(numbers_of_alternate_alleles)
        counts = df_numbers_of_alternate_alleles.apply(pd.Series.value_counts, axis=0, normalize=True).fillna(0)

        try:
            result['missing_freq'] = counts.loc[-1].values.tolist()
        except KeyError:
            result['missing_freq'] = np.zeros(counts.columns.size).tolist()

        df_numbers_of_alternate_alleles_with_nan = df_numbers_of_alternate_alleles.replace(-1, np.nan)
        counts_without_missing = df_numbers_of_alternate_alleles_with_nan.apply(pd.Series.value_counts, axis=0, normalize=True, dropna=True).fillna(0)
        counts_without_missing.index = counts_without_missing.index.astype(int, copy=False)

        try:
            result['heterozygosity_freq'] = counts_without_missing.loc[1].values.tolist()
        except KeyError:
            result['heterozygosity_freq'] = np.zeros(counts_without_missing.columns.size).tolist()

        return result


    def apply_variant_filter_settings(self, fs, numbers_of_alternate_alleles, _slice_variant_calls):
        variants_summary_stats = self.calc_variants_summary_stats(numbers_of_alternate_alleles)

        if 'QUAL' in self.available_variants_metadata:
            sliced_qual = self.variants_qual.get_basic_selection(_slice_variant_calls)
            variants_summary_stats['vcf_qual'] = sliced_qual.tolist()
        
        variants_summary_stats['positions_indices'] = list(range(_slice_variant_calls.start, _slice_variant_calls.stop))
        df = pd.DataFrame(variants_summary_stats)

        log.debug(df)

        if 'filterByMaf' in fs and fs['filterByMaf'] == True:
            df = df[ df['maf'].between(fs['maf'][0], fs['maf'][1]) ]

        if 'filterByMissingFreq' in fs and fs['filterByMissingFreq'] == True:
            df = df[ df['missing_freq'].between(fs['missingFreq'][0], fs['missingFreq'][1]) ]

        if 'filterByHeteroFreq' in fs and fs['filterByHeteroFreq'] == True:
            df = df[ df['heterozygosity_freq'].between(fs['heteroFreq'][0], fs['heteroFreq'][1]) ]

        if 'filterByVcfQual' in fs and fs['filterByVcfQual'] == True and 'QUAL' in self.available_variants_metadata:
            df = df[ df['vcf_qual'].between(fs['vcfQual'][0], fs['vcfQual'][1]) ]

        if numbers_of_alternate_alleles[:, df.index.values].shape[1] >= 2:
            numbers_of_alternate_alleles = numbers_of_alternate_alleles[:, df.index.values]

        return numbers_of_alternate_alleles, df['positions_indices'].values


    def get_slice_of_variant_calls(self, chrom, startpos=None, endpos=None, count=None, samples=None, variant_filter_settings=None):

        lookup_type_start = False
        lookup_type_end = False

        samples_mask, samples_selected_mapped = self.get_samples_mask(samples)

        if count is None:
            location_start, lookup_type_start = self.get_posidx_by_genome_coordinate(chrom, startpos)
            location_end, lookup_type_end = self.get_posidx_by_genome_coordinate(chrom, endpos)
            location_end = location_end + 1
        else:
            if startpos is not None:
                # Get start coordinate (allows automatic position fuzzy search if coordinate does not exist in the variant matrix!)
                location_start, lookup_type_start = self.get_posidx_by_genome_coordinate(chrom, startpos)

                # calculate location_end from start and count
                location_end = location_start + count
                

            if endpos is not None:
                # Get start coordinate (allows automatic position fuzzy search if coordinate does not exist in the variant matrix!)
                location_end, lookup_type_end = self.get_posidx_by_genome_coordinate(chrom, endpos)
                location_end = location_end + 1

                # calculate location_start from end and count
                location_start = location_end - count

                if location_start < 0:
                    location_start = 0
                    location_end = location_start + count


        # create slice() object for later going into get_orthogonal_selection()
        slice_variant_calls = slice(location_start, location_end, None)

        positions = self.pos[slice_variant_calls]

        # get the variant slice from Zarr dataset
        sliced_variant_calls = self.calldata.get_orthogonal_selection((slice_variant_calls, samples_mask))   # samples_mask

        # Transform each genotype call into the number of non-reference alleles and then transpose it via .T to row-major order
        numbers_of_alternate_alleles = self.count_alternate_alleles(sliced_variant_calls)

        filtered_positions_indices = np.asarray(range(slice_variant_calls.start, slice_variant_calls.stop)) #None
        if variant_filter_settings is not None:
            numbers_of_alternate_alleles, filtered_positions_indices = self.apply_variant_filter_settings(variant_filter_settings, numbers_of_alternate_alleles, slice_variant_calls)

        stats = {
            'number_of_variants_in_window': int(positions.shape[0]),
            'number_of_variants_in_window_filtered': int(numbers_of_alternate_alleles.shape[1]),
            'startpos': int(positions[0]),
            'endpos': int(positions[-1]),
            'lookup_type_start': str(lookup_type_start),
            'lookup_type_end': str(lookup_type_end)
        }

        result = {
            'numbers_of_alternate_alleles': numbers_of_alternate_alleles,
            'slice_variant_calls': slice_variant_calls,
            'sliced_variant_calls': sliced_variant_calls,
            'samples_mask': samples_mask,
            'samples_selected_mapped': samples_selected_mapped,
            'positions': positions,
            'filtered_positions_indices': filtered_positions_indices,
            'location_start': location_start,
            'location_end': location_end,
            'stats': stats
        }

        return SimpleNamespace(**result)



import os, sys

sys.setrecursionlimit(1000000)

from pprint import pprint

import click
import yaml
import gzip
import allel
import zarr
import pandas as pd
import numcodecs
from waitress import serve

from divbrowse import log


from divbrowse import __version__ as DIVBROWSE_VERSION

def get_config_skeleton():
    path_config_skeleton = os.path.join(os.path.dirname(__file__), 'divbrowse.config.yml.skeleton')
    try:
        with open(path_config_skeleton) as config_file:
            config = yaml.full_load(config_file)
    except FileNotFoundError:
        log.error('Divbrowse config file `divbrowse.config.yml` not found in current directory!')
        exit(1)

    return config


def get_chromosomes(path_zarr):
    callset = zarr.open_group(path_zarr, mode='r')
    chromosomes = pd.unique(callset['variants/CHROM'][:]).tolist()
    return chromosomes



@click.group()
@click.version_option(prog_name='DivBrowse', version=DIVBROWSE_VERSION)
def main():
    """This is the DivBrowse CLI"""
    pass


@click.command()
def calcsumstats():

    click.echo('Starting calculation of variant summary statistics...')
    log.info('Starting calculation of variant summary statistics...')

    from divbrowse.lib.annotation_data import AnnotationData
    from divbrowse.lib.genotype_data import GenotypeData

    try:
        with open('divbrowse.config.yml') as config_file:
            config = yaml.full_load(config_file)
    except FileNotFoundError:
        log.error('Divbrowse config file `divbrowse.config.yml` not found in current directory!')
        exit(1)

    gd = GenotypeData(config)
    ad = AnnotationData(config, gd)

    vcf_header_lines = gd.get_vcf_header()
    print(vcf_header_lines)





@click.command()
@click.option('--path-vcf', help='Full path to to VCF file that should be converted to a Zarr archive')
@click.option('--path-zarr', help='Full path where to save the Zarr archive')
def vcf2zarr(path_vcf: str, path_zarr: str):

    if path_vcf == None:
        vcf_files = []
        path = os.getcwd()

        with os.scandir(path) as entries:
            for entry in entries:
                if entry.is_file() and entry.name.lower().endswith(('.vcf', '.vcf.gz')):
                    vcf_files.append(entry.path)
        
        click.secho('The following VCF files have been found in current working directory. Please choose the one you want to convert to the Zarr format:', fg='yellow')
        i = 0
        for _vcf_file in vcf_files:
            click.secho('['+str(i)+'] '+os.path.basename(_vcf_file), fg='yellow')
            i = i + 1

        selected_vcf_number = click.prompt(click.style('Please enter the number of the VCF file', fg='yellow'), type=int)

        if selected_vcf_number >= len(vcf_files):
            log.error('The given number for a VCF file is not valid.')
            exit(1)

        path_vcf = str(vcf_files[selected_vcf_number])
        click.secho('You have selected the following VCF file:: '+path_vcf, fg='yellow')

    if path_zarr == None:
        path_zarr = path_vcf + '.zarr'


    vcf_header_lines = []
    if path_vcf.endswith('gz'):
        vcf_file = gzip.open(path_vcf, mode='rt', encoding='utf-8')
    else:
        vcf_file = open(path_vcf, mode='r', buffering=0, encoding='utf-8')

    vcf_line = vcf_file.readline()
    while vcf_line:
        if vcf_line.startswith('##'):
            vcf_header_lines.append(vcf_line)
            vcf_line = vcf_file.readline()
        else:
            break

    with open("____vcf_export_header_lines____.vcf", "w") as vcf_header_lines_file:
        vcf_header_lines_file.write("".join(vcf_header_lines))


    try:
        allel.vcf_to_zarr(path_vcf, path_zarr, group='/', fields='*', exclude_fields=['variants/numalt', 'variants/altlen', 'variants/is_snp'], log=sys.stdout, compressor=numcodecs.Blosc(cname='zstd', clevel=5, shuffle=False))
    except ValueError as error_msg:
        log.error(error_msg)


    click.secho('Conversion to Zarr finished.', fg='green')
    click.secho('The Zarr archive was saved to this path: '+path_zarr, fg='green')



@click.command()
@click.option('--host', default='0.0.0.0', help='IP address to bind the DivBrowse server to', show_default=True)
@click.option('--port', default='8080', help='Port number to bind the DivBrowse server to', show_default=True)
@click.option('--infer-config', is_flag=True, help='If set: infer a basic configuration from the provided VCF and GFF/GFF3 files and do not use an existing `divbrowse.config.yml`')
@click.option('--save-config', type=click.Path(file_okay=True, writable=True), help='Save the inferred configuration as a YAML file. Please provide a relative or absolute path.')
def start(host: str, port: str, infer_config: bool, save_config):
    from divbrowse.server import create_app

    if infer_config:

        vcf_files = []
        path = os.getcwd()

        with os.scandir(path) as entries:
            for entry in entries:
                if entry.is_file() and entry.name.lower().endswith(('.vcf', '.vcf.gz')):
                    vcf_files.append(entry.path)
        
        click.secho('The following VCF files have been found. Please choose the one you want to visualize with DivBrowse:', fg='yellow')
        i = 0
        for _vcf_file in vcf_files:
            click.secho('['+str(i)+'] '+os.path.basename(_vcf_file), fg='yellow')
            i = i + 1

        selected_vcf_number = click.prompt(click.style('Please enter the number of the VCF file', fg='yellow'), type=int)

        if selected_vcf_number >= len(vcf_files):
            log.error('The given number for a VCF file is not valid.')
            exit(1)

        click.secho('Your VCF choice: '+str(vcf_files[selected_vcf_number]), fg='yellow')
        selected_vcf = vcf_files[selected_vcf_number]

        path_zarr = selected_vcf + '.zarr'
        if not os.path.isdir(path_zarr):
            allel.vcf_to_zarr(selected_vcf, path_zarr, group='/', fields='*', log=sys.stdout, compressor=numcodecs.Blosc(cname='zstd', clevel=1, shuffle=False)) # cname='zstd'


        gff3_files = []
        with os.scandir(path) as entries:
            for entry in entries:
                if entry.is_file() and entry.name.lower().endswith(('.gff', '.gff3')):
                    gff3_files.append(entry.path)


        click.secho('The following GFF/GFF3 files have been found. Please choose the one you want to visualize with DivBrowse:', fg='yellow')
        i = 0
        for _gff3_file in gff3_files:
            click.secho('['+str(i)+'] '+os.path.basename(_gff3_file), fg='yellow')

        selected_gff3_number = click.prompt(click.style('Please enter the number of the GFF/GFF3 file', fg='yellow'), type=int)

        if selected_gff3_number >= len(gff3_files):
            log.error('The given number for a GFF/GFF3 file is not valid.')
            exit(1)

        click.secho('Your GFF/GFF3 choice: '+str(gff3_files[selected_gff3_number]), fg='yellow')
        selected_gff3 = gff3_files[selected_gff3_number]


        chromosomes_vcf = get_chromosomes(path_zarr)
        chromosome_labels = dict(zip(chromosomes_vcf, chromosomes_vcf))

        centromeres_positions = {k:'0' for k, v in chromosome_labels.items()}

        genes = allel.gff3_to_dataframe(selected_gff3)
        chromosomes_gff = genes['seqid'].unique().tolist()
        gff3_chromosome_labels = dict(zip(chromosomes_vcf, chromosomes_gff))

        config_runtime = get_config_skeleton()

        config_runtime['datadir'] = path+'/'
        config_runtime['variants']['zarr_dir'] = os.path.basename(path_zarr)
        config_runtime['gff3']['filename'] = os.path.basename(gff3_files[0])
        config_runtime['gff3']['feature_type_with_description'] = 'gene'
        config_runtime['gff3']['main_feature_types_for_genes_track'] = ['gene']
        config_runtime['chromosome_labels'] = chromosome_labels
        config_runtime['centromeres_positions'] = centromeres_positions
        config_runtime['gff3_chromosome_labels'] = gff3_chromosome_labels

        if save_config is not None:
            with open(save_config, 'w') as yaml_file:
                yaml.dump(config_runtime, yaml_file, default_flow_style=False)

    click.secho('Starting DivBrowse server...', fg='green', bold=True)

    if host == '0.0.0.0':
        import socket
        url = 'http://'+socket.gethostname()+':'+str(port)+'/index.html'
    else:
        url = 'http://'+str(host)+':'+str(port)+'/index.html'

    click.secho('DivBrowse should be available under the following URL: '+str(url), fg='green', bold=True)

    if infer_config:
        app = create_app(config_runtime=config_runtime)
    else:
        app = create_app()

    serve(app, host=host, port=int(port))





main.add_command(vcf2zarr)
main.add_command(start)
main.add_command(calcsumstats)

if __name__ == '__main__':
    main()


__version__ = '0.1.0'

import json
from timeit import default_timer as timer

from flask import Flask, Response, jsonify, request

import allel
import numpy as np
import pandas as pd
from bioblend.galaxy import GalaxyInstance
from bioblend.galaxy.tools.inputs import inputs
from sklearn.metrics import pairwise_distances
import yaml

from divbrowse import log
from divbrowse.lib.annotation_data import AnnotationData
from divbrowse.lib.genotype_data import (GenotypeData, calc_pca_for_slice_of_variant_calls,
    calc_umap_for_slice_of_variant_calls, impute_with_mean)

from divbrowse.lib.utils import ApiError
from divbrowse.lib.utils import StrictEncoder


def create_app(filename_config_yaml = 'divbrowse.config.yml', config_runtime=None):
    """Factory method to create and return a wsgi-compliant Flask app instance"""

    app = Flask(__name__, static_url_path='', static_folder='static')

    app.json_encoder = StrictEncoder

    if config_runtime is not None:
        log.info('Using runtime config')
        config = config_runtime
    else:

        try:
            with open(filename_config_yaml) as config_file:
                config = yaml.full_load(config_file)
        except FileNotFoundError:
            log.error('Divbrowse config file `divbrowse.config.yml` not found in current directory!')
            exit(1)


    log.info('Instanciate GenotypeData classes')
    gd = GenotypeData(config)

    log.info('Instanciate AnnotationData classes')
    ad = AnnotationData(config, gd)


    def process_request_vars(vars):

        processed = {
            'chrom': vars['chrom'],
        }

        try:

            if 'samples' in vars:
                # conversion to Python list from stringified JSON is necessary for form-data POST requests of vcf_export
                if type(vars['samples']) is str:
                    vars['samples'] = json.loads(vars['samples'])

                samples, unmapable_sample_ids = gd.map_input_sample_ids_to_vcf_sample_ids(vars['samples'])

                if len(unmapable_sample_ids) > 0:
                    raise ApiError('The following sample-IDs could not be resolved: '+', '.join(unmapable_sample_ids))

                processed['samples'] = samples

            processed['count'] = None
            if 'count' in vars:
                processed['count'] = int(vars['count'])

            processed['startpos'] = None
            if 'startpos' in vars:
                processed['startpos'] = int(vars['startpos'])

            processed['endpos'] = None
            if 'endpos' in vars:
                processed['endpos'] = int(vars['endpos'])

            processed['variant_filter_settings'] = None
            if 'variant_filter_settings' in vars:
                # conversion to Python list from stringified JSON is necessary for form-data POST requests of vcf_export
                if type(vars['variant_filter_settings']) is str:
                    vars['variant_filter_settings'] = json.loads(vars['variant_filter_settings'])
                processed['variant_filter_settings'] = vars['variant_filter_settings']

            return processed

        except KeyError:
            raise ApiError('Some input data is missing.')




    @app.route("/snp_window_summary", methods = ['GET', 'POST', 'OPTIONS'])
    def __snp_window_summary():

        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            return 'ERROR'

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        return jsonify(_result.stats)



    @app.route("/pca", methods = ['GET', 'POST', 'OPTIONS'])
    def __pca():

        payload = request.get_json(silent=True)

        if request.method == 'POST':
            input = process_request_vars(payload)
        else:
            return 'ERROR'

        umap_n_neighbors = int(payload['umap_n_neighbors'])

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        pca_result, pca_explained_variance = calc_pca_for_slice_of_variant_calls(_result.numbers_of_alternate_alleles, _result.samples_selected_mapped)

        umap_result = calc_umap_for_slice_of_variant_calls(
            slice_of_variant_calls = _result.numbers_of_alternate_alleles, 
            samples_selected = _result.samples_selected_mapped,
            n_neighbors = umap_n_neighbors
        )

        result = {
            'pca_result': pca_result.tolist(),
            'pca_explained_variance': pca_explained_variance.tolist(),
            'umap_result': umap_result.tolist(),
        }

        return jsonify(result)




    @app.route("/variant_calls", methods = ['GET', 'POST', 'OPTIONS'])
    def __variant_calls():

        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the variant matrix.'
            })
        
        slice = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            count = input['count'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        calls = {}
        if slice.sliced_variant_calls.ndim == 2:
            slice.sliced_variant_calls = slice.sliced_variant_calls.T # transpose GenotypeArray so that samples are in the 1st dimension and not the variant-calls

        if slice.sliced_variant_calls.ndim == 3:
            slice.sliced_variant_calls = slice.sliced_variant_calls.transpose(1, 0, 2) # transpose GenotypeArray so that samples are in the 1st dimension and not the variant-calls

        i = 0
        
        for sample in slice.samples_selected_mapped:
            calls[sample] = slice.sliced_variant_calls[i].tolist()
            i += 1

        result = {
            'calls': calls,
            'calls_metadata': {}
        }

        # get DP values
        if 'DP' in gd.available_calldata:
            DP_values = {}
            sliced_DP = gd.callset['calldata/DP'].get_orthogonal_selection((slice.slice_variant_calls, slice.samples_mask)).T
            DP_values = {}
            i = 0
            for sample in  slice.samples_selected_mapped:
                DP_values[sample] = sliced_DP[i].tolist()
                i += 1
            result['calls_metadata']['dp'] = DP_values


        # get DV values
        if 'DV' in gd.available_calldata:
            DV_values = {}
            sliced_DV = gd.callset['calldata/DV'].get_orthogonal_selection((slice.slice_variant_calls, slice.samples_mask)).T
            DV_values = {}
            i = 0
            for sample in  slice.samples_selected_mapped:
                DV_values[sample] = sliced_DV[i].tolist()
                i += 1
            result['calls_metadata']['dv'] = DV_values

        return jsonify(result)


    @app.route("/variants", methods = ['GET', 'POST', 'OPTIONS'])
    def __variants():

        start_all = timer()

        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the variant matrix.'
            })
        
        log.debug("==== 0 => calculation time: %f", timer() - start_all)

        start = timer()

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            count = input['count'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        numbers_of_alternate_alleles = _result.numbers_of_alternate_alleles
        _slice_variant_calls = _result.slice_variant_calls
        _sliced_variant_calls = _result.sliced_variant_calls
        samples_mask = _result.samples_mask
        samples_selected_mapped = _result.samples_selected_mapped
        positions = _result.positions
        location_start = _result.location_start
        location_end = _result.location_end

        # calculate some allel stats like MAF etc.
        variants_summary_stats = gd.calc_variants_summary_stats(numbers_of_alternate_alleles)

        # impute with mean
        imputed = impute_with_mean(numbers_of_alternate_alleles)
        ref_vec = np.zeros(numbers_of_alternate_alleles.shape[1]).reshape(1, numbers_of_alternate_alleles.shape[1])
        log.debug("==== gd.get_variant_matrix() + gd.calc_variants_summary_stats() + impute_with_mean() => calculation time: %f", timer() - start)

        start = timer()
        distances = pairwise_distances(imputed, ref_vec, n_jobs=1, metric='hamming')
        distances = distances * numbers_of_alternate_alleles.shape[1]
        distances = distances.astype(np.int16);
        sample_ids = np.array(samples_selected_mapped).reshape(gd.samples[samples_mask].shape[0], 1)
        distances_combined = np.concatenate((sample_ids, distances), axis=1)
        log.debug("==== pairwise_distances() calculation time: %f", timer() - start)

        # Get the reference nucleotides (as letters ATCG)
        sliced_reference = gd.reference_allele[_slice_variant_calls]

        # Get the alternate nucleotides (as letters ATCG)
        sliced_alternates = gd.alternate_alleles[_slice_variant_calls]

        '''
        variants_samples = {}
        calls = {}
        if _sliced_variant_calls.ndim == 2:
            _sliced_variant_calls = _sliced_variant_calls.T # transpose GenotypeArray so that samples are in the 1st dimension and not the variant-calls

        if _sliced_variant_calls.ndim == 3:
            _sliced_variant_calls = _sliced_variant_calls.transpose(1, 0, 2) # transpose GenotypeArray so that samples are in the 1st dimension and not the variant-calls

        i = 0
        
        for sample in samples_selected_mapped:
            variants_samples[sample] = numbers_of_alternate_alleles[i].tolist()
            calls[sample] = _sliced_variant_calls[i].tolist()
            i += 1
        '''

        start = timer()

        result = {
            'coordinate_first': int(gd.pos[location_start]),
            'coordinate_last': int(gd.pos[location_end-1]),
            'coordinate_first_next': int(gd.pos[location_end]),
            'coordinate_last_prev': int(gd.pos[location_start-1]),
            'coordinate_first_chromosome': gd.chrom[location_start],
            'coordinate_last_chromosome': gd.chrom[location_end],
            'variants_coordinates': positions.tolist(),
            #'variants': variants_samples,
            #'calls': calls,
            'reference': sliced_reference.tolist(),
            'alternates': sliced_alternates.tolist(),
            'hamming_distances_to_reference': distances_combined.tolist()
        }

        '''
        # get DP values
        if 'DP' in gd.available_calldata:
            DP_values = {}
            sliced_DP = gd.callset['calldata/DP'].get_orthogonal_selection((_slice_variant_calls, samples_mask)).T
            DP_values = {}
            i = 0
            for sample in samples_selected_mapped:
                DP_values[sample] = sliced_DP[i].tolist()
                i += 1
            result['dp'] = DP_values


        # get DV values
        if 'DV' in gd.available_calldata:
            DV_values = {}
            sliced_DV = gd.callset['calldata/DV'].get_orthogonal_selection((_slice_variant_calls, samples_mask)).T
            DV_values = {}
            i = 0
            for sample in samples_selected_mapped:
                DV_values[sample] = sliced_DV[i].tolist()
                i += 1
            result['dv'] = DV_values
        '''

        #### QUAL #########################
        if 'QUAL' in gd.available_variants_metadata:
            sliced_qual = gd.variants_qual.get_basic_selection(_slice_variant_calls)
            variants_summary_stats['vcf_qual'] = sliced_qual.tolist()



        result['per_snp_stats'] = variants_summary_stats

        #### SNPEFF #######################
        if gd.available['snpeff']:
            sliced_ann = gd.callset['variants/ANN'].get_basic_selection( _slice_variant_calls )
            
            snpeff_variants = {}
            i = 0
            for snpeff_variant_pos in positions.tolist():
                if isinstance(sliced_ann[i], str):
                    snpeff_variants[snpeff_variant_pos] = sliced_ann[i]
                else:
                    snpeff_variants[snpeff_variant_pos] = sliced_ann[i].tolist()
                i += 1

            result['snpeff_variants_coordinates'] = positions.tolist()
            result['snpeff_variants'] = snpeff_variants


        #### GFF3 ###########################################
        if ad.available['gff3']:
            curr_start = int(gd.pos[location_start])
            curr_end = int(gd.pos[location_end-1])

            start = timer()
            genes_within_slice = ad.genes.loc[ ( ad.genes['start'] <= curr_start) & (ad.genes['end'] >= curr_end ) ]
            genes_starting_in_slice = ad.genes.loc[ ( ad.genes['start'] >= curr_start) & (ad.genes['start'] <= curr_end ) ]
            genes_ending_in_slice = ad.genes.loc[ ( ad.genes['end'] >= curr_start) & (ad.genes['end'] <= curr_end ) ]
            genes_all_in_slice = pd.concat([genes_within_slice, genes_starting_in_slice, genes_ending_in_slice]).drop_duplicates().reset_index(drop=True)
            genes_all_in_slice = genes_all_in_slice.loc[ (genes_all_in_slice['seqid'] == ad.chrom_gff3_map[input['chrom']]) ]
            calctime = timer() - start
            log.debug("==== time for genes lookup: %f", calctime)
            result['features'] = genes_all_in_slice.to_dict(orient='records')

            #### Nearest gene ##############################
            nearest_gene = ad.get_nearest_gene_start_pos(input['chrom'], int(gd.pos[location_start]))
            result['nearest_feature'] = nearest_gene.to_dict(orient='records')
        

        log.debug("==== output dict creation => calculation time: %f", timer() - start)
        log.debug("==== ALL /variants => calculation time: %f", timer() - start_all)

        return jsonify(result)



    @app.route("/vcf_export_check", methods = ['GET', 'POST', 'OPTIONS'])
    def __vcf_export_check():
        
        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error_missing_chromosome', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the variant matrix.'
            })

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            count = input['count'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        if _result.stats['number_of_variants_in_window_filtered'] > 5000:
            return jsonify({
                'success': False, 
                'status': 'error_snp_window_too_big', 
                'message': 'The requested genomic window size is bigger than 5000 variants and is therefore too big. Please decrease the window size to not exceed 5000 variants.'
            })

        return jsonify({
            'success': True, 
            'status': 'export_possible',
            'message': _result.stats['number_of_variants_in_window_filtered']
        })


    @app.route("/vcf_export", methods = ['GET', 'POST', 'OPTIONS'])
    def __vcf_export():

        if request.method == 'POST':
            input = process_request_vars(request.form.to_dict())
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error_missing_chromosome', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the variant matrix.'
            })

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        numbers_of_alternate_alleles = _result.numbers_of_alternate_alleles
        _slice_variant_calls = _result.slice_variant_calls
        _sliced_variant_calls = _result.sliced_variant_calls
        samples_mask = _result.samples_mask
        samples_selected_mapped = _result.samples_selected_mapped
        positions = _result.positions
        filtered_positions_indices = _result.filtered_positions_indices
        location_start = _result.location_start
        location_end = _result.location_end

        vcf_lines_header = [
            '##fileformat=VCFv4.0',
            #'##fileDate=20190225',
            #'##source=SeqArray_Format_v1.0',
            #'##reference=Morex v2',
            #'##INFO=<ID=MQ,Number=1,Type=Integer,Description="Average mapping quality">',
            '##FILTER=<ID=PASS,Description="All filters passed">',
            '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
            #'##FORMAT=<ID=DP,Number=.,Type=Integer,Description="Read depth">',
            #'##FORMAT=<ID=DV,Number=.,Type=Integer,Description="Read depth of the alternative allele">'
        ]

        mapped_sample_ids, _ = gd.map_vcf_sample_ids_to_input_sample_ids(gd.samples[samples_mask].astype(str).tolist())
        vcf_line_variants_header = ['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT'] + mapped_sample_ids
        vcf_lines_header.append("\t".join(vcf_line_variants_header)) 

        ref = gd.reference_allele.get_orthogonal_selection( (filtered_positions_indices) )
        alts = gd.alternate_alleles.get_orthogonal_selection( (filtered_positions_indices) )
        qual = gd.callset['variants/QUAL'].get_orthogonal_selection( (filtered_positions_indices) )
        #mq = callset['variants/MQ'].get_orthogonal_selection( (filtered_positions_indices) )

        def generate():
            
            yield "\n".join(vcf_lines_header) + "\n"

            i = 0
            for pos_idx in filtered_positions_indices.tolist():

                vcf_line = [
                    str(input['chrom']),
                    str(gd.pos[pos_idx]),
                    '.',
                    str(ref[i]),
                    ','.join([ _alt for _alt in (alts[i].astype(str).tolist()) if _alt != '' ]), #','.join([ _alt for _alt in (alts[i].astype(str).tolist()) if _alt != '' ]),
                    str(qual[i]),
                    'NA',
                    '', #'MQ='+str(mq[i]),
                    'GT' #'GT:DP:DV'
                ]

                if 'samples' in input:
                    gt_slice = gd.callset['calldata/GT'].get_orthogonal_selection( ([pos_idx], samples_mask, slice(None)) )
                else:
                    gt_slice = gd.callset['calldata/GT'].get_orthogonal_selection( ([pos_idx], slice(None), slice(None)) )
                ga = allel.GenotypeArray(gt_slice)
                gt = ga.to_gt()

                vcf_line = vcf_line + gt[0].astype(str).tolist()
                yield "\t".join(vcf_line)+"\n"
                i = i + 1

        return Response(generate(), mimetype='text/csv', headers={"Content-Disposition":"attachment; filename=custom_export.vcf"})



    @app.route("/gff3_export", methods = ['GET', 'POST', 'OPTIONS'])
    def __gff3_export():

        if request.method == 'POST':
            print(request.form)
            print(request.form.to_dict())
            input = process_request_vars(request.form.to_dict())
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error_missing_chromosome', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the SNP matrix.'
            })

        #df_gff3_slice = self.genes[ (self.genes['seqid'] == 0) & (self.genes['start'] >= 0) & (self.genes['end'] <= 0) ]

        curr_start = input['startpos']
        curr_end = input['endpos']

        genes_within_slice = ad.genes.loc[ ( ad.genes['start'] <= curr_start) & (ad.genes['end'] >= curr_end ) ]
        genes_starting_in_slice = ad.genes.loc[ ( ad.genes['start'] >= curr_start) & (ad.genes['start'] <= curr_end ) ]
        genes_ending_in_slice = ad.genes.loc[ ( ad.genes['end'] >= curr_start) & (ad.genes['end'] <= curr_end ) ]
        genes_all_in_slice = pd.concat([genes_within_slice, genes_starting_in_slice, genes_ending_in_slice]).drop_duplicates().reset_index(drop=True)
        genes_all_in_slice = genes_all_in_slice.loc[ (genes_all_in_slice['seqid'] == ad.chrom_gff3_map[input['chrom']]) ]

        genes_all_in_slice = genes_all_in_slice.sort_values('start')

        print(genes_all_in_slice)
        
        '''
        key_confidence = 'primary_confidence_class'
        if config['gff3']['key_confidence']:
            key_confidence = str(config['gff3']['key_confidence'])

        key_ontology = 'Ontology_term'
        if config['gff3']['key_ontology']:
            key_ontology = str(config['gff3']['key_ontology'])
        '''


        def generate():

            for index, row in genes_all_in_slice.iterrows():

                gff3_attributes = []
                if row['ID'] != '.':
                    gff3_attributes.append('ID='+str(row['ID']))

                if row['Parent'] != '.':
                    gff3_attributes.append('Parent='+str(row['Parent']))

                if row['description'] != '.':
                    gff3_attributes.append('description='+str(row['description']))

                if row['Ontology_term'] != '.':
                    gff3_attributes.append('Ontology_term='+str(row['Ontology_term']))

                if row['primary_confidence_class'] != '.':
                    gff3_attributes.append('primary_confidence_class='+str(row['primary_confidence_class']))

                _score = str(row['score'])
                _phase = str(row['phase'])

                gff3_line = [
                    str(row['seqid']),
                    str(row['source']),
                    str(row['type']),
                    str(row['start']),
                    str(row['end']),
                    _score if _score != '-1' else '.',
                    str(row['strand']),
                    _phase if _phase != '-1' else '.',
                    ';'.join(gff3_attributes)
                ]

                yield "\t".join(gff3_line)+"\n"

        return Response(generate(), mimetype='text/csv', headers={"Content-Disposition":"attachment; filename=custom_export.gff3"})





    @app.route("/blast", methods = ['GET', 'POST', 'OPTIONS'])
    def __blast():

        if request.method != 'POST':
            return ''

        if config['blast']['active'] is not True:
            return 'BLAST is not allowed'

        gi = GalaxyInstance(url = config['blast']['galaxy_server_url'], email = config['blast']['galaxy_user'], password = config['blast']['galaxy_pass'])

        json_request_vars = request.get_json(force=True, silent=True)

        blast_types = {
            'ncbi_blastn_wrapper_barley': 'megablast',
            'ncbi_tblastn_wrapper_barley': 'tblastn-fast'
        }

        blast_parameters = {
            'query': str(json_request_vars['query']),
            'database': str(config['blast']['blast_database']),
            #'type': config['blast']['blast_type'],
            'type': blast_types[str(json_request_vars['blast_type'])],
            #'galaxy_tool_id': config['blast']['galaxy_tool_id']
            'galaxy_tool_id': str(json_request_vars['blast_type'])
        }

        histories = gi.histories.get_histories()
        history_id = histories[0]['id']

        paste_content_result = gi.tools.paste_content(blast_parameters['query'], history_id, file_name='blast_query.fasta')
        paste_content_dataset_id = paste_content_result['outputs'][0]['id']

        tool_inputs = (
            inputs().set_dataset_param("query", paste_content_dataset_id, src='hda')
            .set_param('db_opts', 'db')
            .set_param('db_opts|database', [blast_parameters['database']])
            .set_param('blast_type', blast_parameters['type'])
        )
        run_tool_result = gi.tools.run_tool(history_id, blast_parameters['galaxy_tool_id'], tool_inputs)
        result_dataset_id = run_tool_result['outputs'][0]['id']

        blast_chromosome_mapping = config['blast']['blast_result_to_vcf_chromosome_mapping']

        blast_result_dataset = gi.datasets.download_dataset(result_dataset_id)
        blast_result_json = []

        blast_result_lines = blast_result_dataset.decode('utf-8').split("\n")
        for line in blast_result_lines:
            if line != "":
                line_parts = line.split("\t")
                _chromosome_vcf = str(blast_chromosome_mapping[line_parts[1]])
                _single_blast_hit = {
                    'chromosome': _chromosome_vcf,
                    'percentage_of_identical_matches': line_parts[2],
                    'alignment_length': line_parts[3],
                    'number_of_mismatches': line_parts[4],
                    'number_of_gap_openings': line_parts[5],
                    'start_of_alignment_in_query': line_parts[6],
                    'end_of_alignment_in_query': line_parts[7],
                    'start_of_alignment_in_subject': line_parts[8],
                    'end_of_alignment_in_subject': line_parts[9],
                    'e_value': line_parts[10],
                    'bit_score': line_parts[11],
                    'snp_count': gd.count_variants_in_window(_chromosome_vcf, int(line_parts[8]), int(line_parts[9]))
                }
                blast_result_json.append(_single_blast_hit)
        
        return jsonify({
            'success': True,
            'blast_result': str(blast_result_dataset.decode('utf-8')),
            'blast_hits': blast_result_json
        })



    @app.route("/configuration", methods = ['GET', 'POST', 'OPTIONS'])
    def __configuration():

        features = {
            'blast': config['blast']['active'],
            'pca': True,
        }

        samples, _ = gd.map_vcf_sample_ids_to_input_sample_ids(gd.samples.tolist())

        result = {
            'ploidy': gd.ploidy,
            'count_genotypes': len(gd.samples),
            'count_variants': len(gd.pos),
            'count_elements': len(gd.samples) * len(gd.pos),
            'chromosomes': gd.list_of_chromosomes,
            'samples': samples,
            'gff3': ad.metadata_gff3,
            'features': features,
            'dataset_descriptions': dict(config['metadata'])
        }

        return jsonify(result)



    @app.route("/chromosomes", methods = ['GET', 'POST', 'OPTIONS'])
    def __chromosomes():
        return jsonify(gd.list_of_chromosomes)


    @app.route("/samples", methods = ['GET', 'POST', 'OPTIONS'])
    def __samples():
        return jsonify(gd.samples.tolist())


    @app.route("/genes", methods = ['GET', 'POST', 'OPTIONS'])
    def __genes():
        r = Response(response=ad.genes_list_json_dumped, status=200, mimetype="application/json")
        r.headers["Content-Type"] = "application/json; charset=utf-8"
        return r
        


    @app.route("/", methods = ['GET', 'POST', 'OPTIONS'])
    def __home():
        return 'Divbrowse server is running'



    @app.errorhandler(ApiError)
    def handle_api_error(error):
        response = jsonify(error.to_dict())
        response.status_code = error.status_code
        return response



    @app.after_request
    def after_request(response):
        header = response.headers
        header['Access-Control-Allow-Origin'] = '*'
        header['Access-Control-Allow-Headers'] = 'Accept, Accept-CH, Accept-Charset, Accept-Datetime, Accept-Encoding, Accept-Ext, Accept-Features, Accept-Language, Accept-Params, Accept-Ranges, Access-Control-Allow-Credentials, Access-Control-Allow-Headers, Access-Control-Allow-Methods, Access-Control-Allow-Origin, Access-Control-Expose-Headers, Access-Control-Max-Age, Access-Control-Request-Headers, Access-Control-Request-Method, Age, Allow, Alternates, Authentication-Info, Authorization, C-Ext, C-Man, C-Opt, C-PEP, C-PEP-Info, CONNECT, Cache-Control, Compliance, Connection, Content-Base, Content-Disposition, Content-Encoding, Content-ID, Content-Language, Content-Length, Content-Location, Content-MD5, Content-Range, Content-Script-Type, Content-Security-Policy, Content-Style-Type, Content-Transfer-Encoding, Content-Type, Content-Version, Cookie, Cost, DAV, DELETE, DNT, DPR, Date, Default-Style, Delta-Base, Depth, Derived-From, Destination, Differential-ID, Digest, ETag, Expect, Expires, Ext, From, GET, GetProfile, HEAD, HTTP-date, Host, IM, If, If-Match, If-Modified-Since, If-None-Match, If-Range, If-Unmodified-Since, Keep-Alive, Label, Last-Event-ID, Last-Modified, Link, Location, Lock-Token, MIME-Version, Man, Max-Forwards, Media-Range, Message-ID, Meter, Negotiate, Non-Compliance, OPTION, OPTIONS, OWS, Opt, Optional, Ordering-Type, Origin, Overwrite, P3P, PEP, PICS-Label, POST, PUT, Pep-Info, Permanent, Position, Pragma, ProfileObject, Protocol, Protocol-Query, Protocol-Request, Proxy-Authenticate, Proxy-Authentication-Info, Proxy-Authorization, Proxy-Features, Proxy-Instruction, Public, RWS, Range, Referer, Refresh, Resolution-Hint, Resolver-Location, Retry-After, Safe, Sec-Websocket-Extensions, Sec-Websocket-Key, Sec-Websocket-Origin, Sec-Websocket-Protocol, Sec-Websocket-Version, Security-Scheme, Server, Set-Cookie, Set-Cookie2, SetProfile, SoapAction, Status, Status-URI, Strict-Transport-Security, SubOK, Subst, Surrogate-Capability, Surrogate-Control, TCN, TE, TRACE, Timeout, Title, Trailer, Transfer-Encoding, UA-Color, UA-Media, UA-Pixels, UA-Resolution, UA-Windowpixels, URI, Upgrade, User-Agent, Variant-Vary, Vary, Version, Via, Viewport-Width, WWW-Authenticate, Want-Digest, Warning, Width, X-Content-Duration, X-Content-Security-Policy, X-Content-Type-Options, X-CustomHeader, X-DNSPrefetch-Control, X-Forwarded-For, X-Forwarded-Port, X-Forwarded-Proto, X-Frame-Options, X-Modified, X-OTHER, X-PING, X-PINGOTHER, X-Powered-By, X-Requested-With'
        return response

    
    return app


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['divbrowse', 'divbrowse.lib']

package_data = \
{'': ['*'], 'divbrowse': ['tests/data/*']}

install_requires = \
['bioblend>=0.16.0,<0.17.0',
 'click>=8.0.1,<9.0.0',
 'flask>=2.0.1,<3.0.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'pyyaml>=5.4.1,<6.0.0',
 'scikit-allel>=1.3.5,<2.0.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'simplejson>=3.17.3,<4.0.0',
 'tables>=3.6.1,<4.0.0',
 'umap-learn>=0.5.2,<0.6.0',
 'waitress==2.0.0',
 'zarr>=2.8.3,<3.0.0']

extras_require = \
{'docs': ['sphinx>=4.0.2,<5.0.0',
          'sphinx-autoapi>=1.6.0,<2.0.0',
          'sphinx_rtd_theme>=0.5.2,<0.6.0',
          'sphinx-click>=3.0.1,<4.0.0']}

entry_points = \
{'console_scripts': ['divbrowse = divbrowse.cli:main']}

setup_kwargs = {
    'name': 'divbrowse',
    'version': '1.0.0',
    'description': 'A web application for interactive visualization and analysis of genotypic variant matrices',
    'long_description': None,
    'author': 'Patrick König',
    'author_email': 'koenig@ipk-gatersleben.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os
from timeit import default_timer as timer
import json

import allel
import numpy as np
import pandas as pd
import zarr
#from pandarallel import pandarallel
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

#pandarallel.initialize(nb_workers=3, progress_bar=True)


class AnnotationData:

    def __init__(self, config, gd):

        self.config = config
        self.gd = gd
        self.datadir = config['datadir']

        self.available = {
            'gff3': False
        }

        if config['gff3']['filename']:
            self.path_gff3 = self.datadir + config['gff3']['filename']
            if os.path.exists(self.path_gff3):
                self.available['gff3'] = True
            else:
                exit('ERROR: the configured path for the GFF3 file does not exist or is not accessible')

        self.metadata_gff3 = {
            'has_gff3': self.available['gff3'],
            'count_genes': 0,
        }
        self.genes_list = []

        self.genes_list_json_dumped = json.dumps({'genes': False})

        if self.available['gff3']:
            self._load_gff3_data()


    def _load_gff3_data(self):

        gff3_chromosome_labels = self.config['gff3_chromosome_labels']
        self.metadata_gff3['gff3_to_vcf_chromosome_mapping'] = {str(v): str(k) for k, v in self.config['gff3_chromosome_labels'].items()}

        self.chrom_gff3_map = { str(key): str(value) for (key, value) in gff3_chromosome_labels.items()}

        key_confidence = 'primary_confidence_class'
        if self.config['gff3']['key_confidence'] and self.config['gff3']['key_confidence'] is not False:
            key_confidence = str(self.config['gff3']['key_confidence'])

        key_ontology = 'Ontology_term'
        if self.config['gff3']['key_ontology'] and self.config['gff3']['key_ontology'] is not False:
            key_ontology = str(self.config['gff3']['key_ontology'])

        gff3_attributes = ['ID', key_confidence, 'Parent', 'description', key_ontology]

        additional_attributes_keys = []
        if self.config['gff3']['additional_attributes_keys']:
            additional_attributes_keys = str(self.config['gff3']['additional_attributes_keys']).split(',')
            gff3_attributes.extend(additional_attributes_keys)

        self.genes = allel.gff3_to_dataframe(self.path_gff3, attributes=gff3_attributes)
        self.genes.rename(columns={key_confidence: "primary_confidence_class", key_ontology: "Ontology_term"}, inplace=True)

        genes_only = self.genes.loc[(self.genes['type'] == 'gene')] # gene or transcript
        
        self.metadata_gff3['count_genes'] = int(len(genes_only.index))
        self.metadata_gff3.update(dict(self.config['gff3']))

        # list of genes with descriptions and start+end positions
        #genes_with_descriptions = self.genes.loc[(self.genes['type'] == self.config['gff3']['feature_type_with_description'])]

        genes_with_descriptions = self.genes.loc[ self.genes['type'].isin(self.config['gff3']['main_feature_types_for_genes_track']) ]
        #genes_with_descriptions = genes_with_descriptions.loc[ genes_with_descriptions['type'].isin(self.config['gff3']['main_feature_types_for_genes_track']) ]
        genes_with_descriptions = genes_with_descriptions.loc[ genes_with_descriptions['seqid'].isin(list(self.chrom_gff3_map.values())) ]


        def count_exon_variants(row):
            _chromosome_gff3 = row['seqid']
            _chromosome_vcf = self.metadata_gff3['gff3_to_vcf_chromosome_mapping'][ _chromosome_gff3 ]
            number_of_variants = self.gd.count_variants_in_window(str(_chromosome_vcf), row['start'], row['end'])
            return number_of_variants

        def count_genic_variants(row):
            print(str(row['seqid'])+' / '+str(row['start']))
            _chromosome_gff3 = row['seqid']
            _chromosome_vcf = self.metadata_gff3['gff3_to_vcf_chromosome_mapping'][ _chromosome_gff3 ]
            number_of_variants = self.gd.count_variants_in_window(str(_chromosome_vcf), row['start'], row['end'])
            exons = self.genes.loc[(self.genes['seqid'] == row['seqid']) &(self.genes['start'] >= row['start']) & (self.genes['end'] <= row['end']) & (self.genes['type'] == 'exon')]
            number_of_exon_variants = exons.apply(count_exon_variants, axis=1)
            return pd.Series( [number_of_variants, number_of_exon_variants.sum() ], index=['number_of_variants', 'number_of_exon_variants'])


        if self.config['gff3']['count_exon_variants'] is True:

            geneStatsCacheFilename = self.datadir+'____gene_stats_.hdf5'
            try:
                gene_list = pd.read_hdf(geneStatsCacheFilename, key='s')
                print("++++ Loaded Pandas Dataframe for gene stats")

                genes_number_of_variants = gene_list[ ['number_of_variants', 'number_of_exon_variants'] ].copy()
                merged = pd.concat([genes_with_descriptions, genes_number_of_variants], axis=1)
                #gene_list = merged[ ['ID', 'seqid', 'start', 'end', 'primary_confidence_class', 'description', 'Ontology_term', 'number_of_variants', 'number_of_exon_variants'] ].copy()
                gene_list = merged.copy()

            except FileNotFoundError:
                start = timer()
                
                print("++++ Count variants on genes and exons..........")
                #genes_number_of_variants = genes_with_descriptions.parallel_apply(count_genic_variants, axis=1, result_type='expand')
                #genes_with_descriptions = genes_with_descriptions.iloc[0:50]
                genes_number_of_variants = genes_with_descriptions.apply(count_genic_variants, axis=1, result_type='expand')
                print("==== genes_with_descriptions.parallel_apply() calculation time: ", timer() - start)
                merged = pd.concat([genes_with_descriptions, genes_number_of_variants], axis=1)

                #gene_list = merged[ ['ID', 'seqid', 'start', 'end', 'primary_confidence_class', 'description', 'number_of_variants', 'number_of_exon_variants'] ].copy()
                gene_list = merged.copy()
                gene_list.to_hdf(geneStatsCacheFilename, key='s', mode='w', complevel=5, complib='blosc:zstd')

        else:
            gene_list = genes_with_descriptions


        genes_list = gene_list.to_dict('split')

        genes_grouped_by_seqid = genes_only.groupby(self.genes.seqid)
        list_chrom_gff3 = pd.unique(genes_only['seqid']).tolist()
        genes_start_positions = {}
        for _chr in list_chrom_gff3:
            _curr_group = genes_grouped_by_seqid.get_group(_chr)
            genes_start_positions[_chr] = pd.DataFrame(_curr_group['start'].drop_duplicates(keep='first'))
            genes_start_positions[_chr].set_index('start', drop=False, inplace=True)
            genes_start_positions[_chr] = genes_start_positions[_chr].sort_index()

        self.genes_list = genes_list
        self.genes_list_json_dumped = json.dumps({'genes': genes_list})
        self.genes_grouped_by_seqid = genes_grouped_by_seqid
        self.genes_start_positions = genes_start_positions


    def get_nearest_gene_start_pos(self, chrom, pos):
        seqid = self.chrom_gff3_map[chrom]
        nearest = self.genes_start_positions[seqid].index.get_loc(pos, method='nearest')
        start_pos = int(self.genes_start_positions[seqid].iloc[nearest].start)
        _genes = self.genes_grouped_by_seqid.get_group(seqid)
        return _genes.loc[(_genes['start'] == start_pos)]



__version__ = '0.1.0'

import json
from timeit import default_timer as timer

from flask import Flask, Response, jsonify, request
from flask_restx import Api, Resource, reqparse, fields

import allel
import numpy as np
import pandas as pd
from bioblend.galaxy import GalaxyInstance
from bioblend.galaxy.tools.inputs import inputs
from sklearn.metrics import pairwise_distances
import yaml

from divbrowse import log
from divbrowse.lib.annotation_data import AnnotationData
from divbrowse.lib.genotype_data import (GenotypeData, calc_pca_for_slice_of_variant_calls,
                               impute_with_mean)

from divbrowse.lib.utils import ApiError
from divbrowse.lib.utils import StrictEncoder


def create_app(filename_config_yaml = 'divbrowse.config.yml', config_runtime=None):
    """Factory method to create and return a wsgi-compliant Flask app instance"""

    app = Flask(__name__, static_url_path='', static_folder='static')
    app.config['RESTX_MASK_SWAGGER'] = False
    app.config['SWAGGER_UI_DOC_EXPANSION'] = 'list'

    api = Api(app,
        version='1.0',
        title='DivBrowse API',
        description='OpenAPI specification of DivBrowse',
        default='divbrowse',
        default_label='divbrowse',
        ordered=True,
        base_url='/',
        doc='/apidoc',
        #doc = '/apidocs'
    )

    #ns = api.namespace('divbrowse', description='DivBrowse API')

    app.json_encoder = StrictEncoder

    if config_runtime is not None:
        log.info('Using runtime config')
        config = config_runtime
    else:

        try:
            with open(filename_config_yaml) as config_file:
                config = yaml.full_load(config_file)
        except FileNotFoundError:
            log.error('Divbrowse config file `divbrowse.config.yml` not found in current directory!')
            exit(1)


    gd = GenotypeData(config)
    ad = AnnotationData(config, gd)


    def process_request_vars(vars):

        processed = {
            'chrom': vars['chrom'],
        }

        try:

            if 'samples' in vars:
                # conversion to Python list from stringified JSON is necessary for form-data POST requests of vcf_export
                if type(vars['samples']) is str:
                    vars['samples'] = json.loads(vars['samples'])

                samples, unmapable_sample_ids = gd.map_input_sample_ids_to_vcf_sample_ids(vars['samples'])

                if len(unmapable_sample_ids) > 0:
                    raise ApiError('The following sample-IDs could not be resolved: '+', '.join(unmapable_sample_ids))

                processed['samples'] = samples

            processed['count'] = None
            if 'count' in vars:
                processed['count'] = int(vars['count'])

            processed['startpos'] = None
            if 'startpos' in vars:
                processed['startpos'] = int(vars['startpos'])

            processed['endpos'] = None
            if 'endpos' in vars:
                processed['endpos'] = int(vars['endpos'])

            processed['variant_filter_settings'] = None
            if 'variant_filter_settings' in vars:
                # conversion to Python list from stringified JSON is necessary for form-data POST requests of vcf_export
                if type(vars['variant_filter_settings']) is str:
                    vars['variant_filter_settings'] = json.loads(vars['variant_filter_settings'])
                processed['variant_filter_settings'] = vars['variant_filter_settings']

            return processed

        except KeyError:
            raise ApiError('Some input data is missing.')




    @app.route("/snp_window_summary", methods = ['GET', 'POST', 'OPTIONS'])
    def __snp_window_summary():

        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            return 'ERROR'

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        return jsonify(_result.stats)



    @app.route("/pca", methods = ['GET', 'POST', 'OPTIONS'])
    def __pca():

        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            return 'ERROR'

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        pca_result = calc_pca_for_slice_of_variant_calls(_result.snps_to_alt, _result.samples_selected_mapped)

        result = {
            'pca_result': pca_result.tolist()
        }

        return jsonify(result)


    model_variant_call = api.model('VariantCall', {
        'sample_id': fields.List(fields.List(fields.Integer))
    })

    model_metadata_integer = api.model('VariantCallMetadataInteger', {
        'sample_id': fields.List(fields.Integer)
    })

    variant_summary_statistics = api.model('VariantSummaryStatistics', {
        'heterozygosity_freq': fields.List(fields.Float),
        'maf': fields.List(fields.Float),
        'missing_freq': fields.List(fields.Float),
        'vcf_qual': fields.List(fields.Float),
    })

    model_genetic_feature = api.model('GeneticFeature', {
        'ID': fields.String,
        'Ontology_term': fields.String,
        'Parent': fields.String,
        'description': fields.String,
        'end': fields.Integer,
        'phase': fields.Integer,
        'primary_confidence_class': fields.String,
        'score': fields.Integer,
        'seqid': fields.String,
        'source': fields.String,
        'start': fields.Integer,
        'strand': fields.String,
        'type': fields.String,
    })

    model_variants = api.model('Variants', {
        #'chromosomes': fields.List(fields.Nested(model_variant_call)),
        'coordinate_first': fields.Integer,
        'coordinate_first_chromosome': fields.String,
        'coordinate_first_next': fields.Integer,
        'coordinate_last': fields.Integer,
        'coordinate_last_chromosome': fields.String,
        'coordinate_last_prev': fields.Integer,
        #'samples': fields.List(fields.String)
        'variants_coordinates': fields.List(fields.Integer),
        'reference': fields.List(fields.String),
        'alternates': fields.List(fields.List(fields.String)),
        'dp': fields.Nested(model_metadata_integer),
        'dv': fields.Nested(model_metadata_integer),
        'variants_calls': fields.Nested(model_variant_call),
        'variants_summary_statistics': fields.Nested(variant_summary_statistics),
        'genetic_features': fields.List(fields.Nested(model_genetic_feature)),
        'nearest_genetic_features': fields.List(fields.Nested(model_genetic_feature)),
    })


    variants_input = api.model('VariantsRequestPayload', {
        'chrom': fields.String(description='Internal ID of the chromosome (as in the VCF)', example='chr1'),
        'startpos': fields.Integer(required=False, description='Start position for the variant calls slice', example=123456),
        'count': fields.Integer(description='Number of variants to be in the slice', example=50),
        'samples': fields.List(fields.String(description='A JSON array of the sample-IDs to be in the slice', example=['HOR_1111', 'HOR_1112']))
    })

    #@app.route("/variants", methods = ['GET', 'POST', 'OPTIONS'])
    #def __variants():

    @api.route('/variants')
    class Variants(Resource):

        @api.expect(variants_input)
        @api.response(200, 'Success', model_variants)
        def post(self):

            start_all = timer()

            if request.method == 'POST':
                input = process_request_vars(request.get_json(silent=True))
            else:
                #raise ApiError('Method not allowed', status_code=405)
                return ''

            if input['chrom'] not in gd.list_chrom:
                return jsonify({
                    'success': False, 
                    'status': 'error', 
                    'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the SNP matrix.'
                })
            
            log.debug("==== 0 => calculation time: %f", timer() - start_all)

            start = timer()

            _result = gd.get_slice_of_variant_calls(
                chrom = input['chrom'],
                startpos = input['startpos'],
                endpos = input['endpos'],
                count = input['count'],
                samples = input['samples'],
                variant_filter_settings = input['variant_filter_settings']
            )

            snps_to_alt = _result.snps_to_alt
            _slice_snps = _result.slice_snps
            _sliced_snps = _result.sliced_snps
            samples_mask = _result.samples_mask
            samples_selected_mapped = _result.samples_selected_mapped
            positions = _result.positions
            location_start = _result.location_start
            location_end = _result.location_end

            # calculate some allel stats like MAF etc.
            variants_summary_stats = gd.calc_variants_summary_stats(snps_to_alt)

            # impute with mean
            imputed = impute_with_mean(snps_to_alt)
            ref_vec = np.zeros(snps_to_alt.shape[1]).reshape(1, snps_to_alt.shape[1])
            log.debug("==== gd.get_slice_of_variant_calls() + gd.calc_variants_summary_stats() + impute_with_mean() => calculation time: %f", timer() - start)

            start = timer()
            distances = pairwise_distances(imputed, ref_vec, n_jobs=1, metric='hamming')
            distances = distances * snps_to_alt.shape[1]
            distances = distances.astype(np.int16);
            sample_ids = np.array(samples_selected_mapped).reshape(gd.samples[samples_mask].shape[0], 1)
            distances_combined = np.concatenate((sample_ids, distances), axis=1)
            log.debug("==== pairwise_distances() calculation time: %f", timer() - start)

            # Get the reference nucleotides (as letters ATCG)
            sliced_reference = gd.reference_allele[_slice_snps]

            # Get the alternate nucleotides (as letters ATCG)
            sliced_alternates = gd.alternate_alleles[_slice_snps]


            variants_samples = {}
            calls = {}
            if _sliced_snps.ndim == 2:
                _sliced_snps = _sliced_snps.T # transpose GenotypeArray so that samples are in the 1st dimension and not the SNP-calls

            if _sliced_snps.ndim == 3:
                _sliced_snps = _sliced_snps.transpose(1, 0, 2) # transpose GenotypeArray so that samples are in the 1st dimension and not the SNP-calls

            i = 0
            
            for sample in samples_selected_mapped:
                variants_samples[sample] = snps_to_alt[i].tolist()
                calls[sample] = _sliced_snps[i].tolist()
                i += 1

            start = timer()

            result = {
                'coordinate_first': int(gd.pos[location_start]),
                'coordinate_last': int(gd.pos[location_end-1]),
                'coordinate_first_next': int(gd.pos[location_end]),
                'coordinate_last_prev': int(gd.pos[location_start-1]),
                'coordinate_first_chromosome': gd.chrom[location_start],
                'coordinate_last_chromosome': gd.chrom[location_end],
                'variants_coordinates': positions.tolist(),
                'variants': variants_samples,
                'calls': calls,
                'reference': sliced_reference.tolist(),
                'alternates': sliced_alternates.tolist(),
                'hamming_distances_to_reference': distances_combined.tolist()
            }

            # get DP values
            if 'DP' in gd.available_calldata:
                DP_values = {}
                sliced_DP = gd.callset['calldata/DP'].get_orthogonal_selection((_slice_snps, samples_mask)).T
                DP_values = {}
                i = 0
                for sample in samples_selected_mapped:
                    DP_values[sample] = sliced_DP[i].tolist()
                    i += 1
                result['dp'] = DP_values


            # get DV values
            if 'DV' in gd.available_calldata:
                DV_values = {}
                sliced_DV = gd.callset['calldata/DV'].get_orthogonal_selection((_slice_snps, samples_mask)).T
                DV_values = {}
                i = 0
                for sample in samples_selected_mapped:
                    DV_values[sample] = sliced_DV[i].tolist()
                    i += 1
                result['dv'] = DV_values


            #### QUAL #########################
            if 'QUAL' in gd.available_variants_metadata:
                sliced_qual = gd.variants_qual.get_basic_selection(_slice_snps)
                variants_summary_stats['vcf_qual'] = sliced_qual.tolist()



            result['per_snp_stats'] = variants_summary_stats

            #### SNPEFF #######################
            if gd.available['snpeff']:
                sliced_ann = gd.callset['variants/ANN'].get_basic_selection( _slice_snps )
                
                snpeff_variants = {}
                i = 0
                for snpeff_variant_pos in positions.tolist():
                    if isinstance(sliced_ann[i], str):
                        snpeff_variants[snpeff_variant_pos] = sliced_ann[i]
                    else:
                        snpeff_variants[snpeff_variant_pos] = sliced_ann[i].tolist()
                    i += 1

                result['snpeff_variants_coordinates'] = positions.tolist()
                result['snpeff_variants'] = snpeff_variants


            #### GFF3 ###########################################
            if ad.available['gff3']:
                curr_start = int(gd.pos[location_start])
                curr_end = int(gd.pos[location_end-1])

                start = timer()
                genes_within_slice = ad.genes.loc[ ( ad.genes['start'] <= curr_start) & (ad.genes['end'] >= curr_end ) ]
                genes_starting_in_slice = ad.genes.loc[ ( ad.genes['start'] >= curr_start) & (ad.genes['start'] <= curr_end ) ]
                genes_ending_in_slice = ad.genes.loc[ ( ad.genes['end'] >= curr_start) & (ad.genes['end'] <= curr_end ) ]
                genes_all_in_slice = pd.concat([genes_within_slice, genes_starting_in_slice, genes_ending_in_slice]).drop_duplicates().reset_index(drop=True)
                genes_all_in_slice = genes_all_in_slice.loc[ (genes_all_in_slice['seqid'] == ad.chrom_gff3_map[input['chrom']]) ]
                calctime = timer() - start
                log.debug("==== time for genes lookup: %f", calctime)
                result['features'] = genes_all_in_slice.to_dict(orient='records')

                #### Nearest gene ##############################
                nearest_gene = ad.get_nearest_gene_start_pos(input['chrom'], int(gd.pos[location_start]))
                result['nearest_feature'] = nearest_gene.to_dict(orient='records')
            

            log.debug("==== output dict creation => calculation time: %f", timer() - start)
            log.debug("==== ALL /variants => calculation time: %f", timer() - start_all)

            return jsonify(result)
            #return result



    @app.route("/vcf_export_check", methods = ['GET', 'POST', 'OPTIONS'])
    def __vcf_export_check():
        
        if request.method == 'POST':
            input = process_request_vars(request.get_json(silent=True))
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error_missing_chromosome', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the SNP matrix.'
            })

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            count = input['count'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        if _result.stats['number_of_variants_in_window_filtered'] > 5000:
            return jsonify({
                'success': False, 
                'status': 'error_snp_window_too_big', 
                'message': 'The requested SNP window size is bigger than 5000 SNPs and is therefore too big. Please decrease the window size to not exceed 5000 SNPs.'
            })

        return jsonify({
            'success': True, 
            'status': 'export_possible',
            'message': _result.stats['number_of_variants_in_window_filtered']
        })


    @app.route("/vcf_export", methods = ['GET', 'POST', 'OPTIONS'])
    def __vcf_export():

        if request.method == 'POST':
            input = process_request_vars(request.form.to_dict())
        else:
            #raise ApiError('Method not allowed', status_code=405)
            return ''

        if input['chrom'] not in gd.list_chrom:
            return jsonify({
                'success': False, 
                'status': 'error_missing_chromosome', 
                'message': 'The provided chromosome number '+str(input['chrom'])+' is not included in the SNP matrix.'
            })

        _result = gd.get_slice_of_variant_calls(
            chrom = input['chrom'],
            startpos = input['startpos'],
            endpos = input['endpos'],
            samples = input['samples'],
            variant_filter_settings = input['variant_filter_settings']
        )

        snps_to_alt = _result.snps_to_alt
        _slice_snps = _result.slice_snps
        _sliced_snps = _result.sliced_snps
        samples_mask = _result.samples_mask
        samples_selected_mapped = _result.samples_selected_mapped
        positions = _result.positions
        filtered_positions_indices = _result.filtered_positions_indices
        location_start = _result.location_start
        location_end = _result.location_end

        vcf_lines_header = [
            '##fileformat=VCFv4.0',
            #'##fileDate=20190225',
            #'##source=SeqArray_Format_v1.0',
            #'##reference=Morex v2',
            #'##INFO=<ID=MQ,Number=1,Type=Integer,Description="Average mapping quality">',
            '##FILTER=<ID=PASS,Description="All filters passed">',
            '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
            #'##FORMAT=<ID=DP,Number=.,Type=Integer,Description="Read depth">',
            #'##FORMAT=<ID=DV,Number=.,Type=Integer,Description="Read depth of the alternative allele">'
        ]

        mapped_sample_ids, _ = gd.map_vcf_sample_ids_to_input_sample_ids(gd.samples[samples_mask].astype(str).tolist())
        vcf_line_variants_header = ['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT'] + mapped_sample_ids
        vcf_lines_header.append("\t".join(vcf_line_variants_header)) 

        ref = gd.reference_allele.get_orthogonal_selection( (filtered_positions_indices) )
        alts = gd.alternate_alleles.get_orthogonal_selection( (filtered_positions_indices) )
        qual = gd.callset['variants/QUAL'].get_orthogonal_selection( (filtered_positions_indices) )
        #mq = callset['variants/MQ'].get_orthogonal_selection( (filtered_positions_indices) )

        def generate():
            
            yield "\n".join(vcf_lines_header) + "\n"

            i = 0
            for pos_idx in filtered_positions_indices.tolist():

                vcf_line = [
                    str(input['chrom']),
                    str(gd.pos[pos_idx]),
                    '.',
                    str(ref[i]),
                    ','.join([ _alt for _alt in (alts[i].astype(str).tolist()) if _alt != '' ]), #','.join([ _alt for _alt in (alts[i].astype(str).tolist()) if _alt != '' ]),
                    str(qual[i]),
                    'NA',
                    '', #'MQ='+str(mq[i]),
                    'GT' #'GT:DP:DV'
                ]

                if 'samples' in input:
                    gt_slice = gd.callset['calldata/GT'].get_orthogonal_selection( ([pos_idx], samples_mask, slice(None)) )
                else:
                    gt_slice = gd.callset['calldata/GT'].get_orthogonal_selection( ([pos_idx], slice(None), slice(None)) )
                ga = allel.GenotypeArray(gt_slice)
                gt = ga.to_gt()

                vcf_line = vcf_line + gt[0].astype(str).tolist()
                yield "\t".join(vcf_line)+"\n"
                i = i + 1

        return Response(generate(), mimetype='text/csv', headers={"Content-Disposition":"attachment; filename=custom_export.vcf"})





    @app.route("/blast", methods = ['GET', 'POST', 'OPTIONS'])
    def __blast():

        if request.method != 'POST':
            return ''

        if config['blast']['active'] is not True:
            return 'BLAST is not allowed'

        gi = GalaxyInstance(url = config['blast']['galaxy_server_url'], email = config['blast']['galaxy_user'], password = config['blast']['galaxy_pass'])

        json_request_vars = request.get_json(force=True, silent=True)

        blast_parameters = {
            'query': str(json_request_vars['query']),
            'database': str(config['blast']['blast_database']),
            'type': config['blast']['blast_type'],
            'galaxy_tool_id': config['blast']['galaxy_tool_id']
        }

        histories = gi.histories.get_histories()
        history_id = histories[0]['id']

        paste_content_result = gi.tools.paste_content(blast_parameters['query'], history_id, file_name='blast_query.fasta')
        paste_content_dataset_id = paste_content_result['outputs'][0]['id']

        tool_inputs = (
            inputs().set_dataset_param("query", paste_content_dataset_id, src='hda')
            .set_param('db_opts', 'db')
            .set_param('db_opts|database', [blast_parameters['database']])
            .set_param('blast_type', blast_parameters['type'])
        )
        run_tool_result = gi.tools.run_tool(history_id, blast_parameters['galaxy_tool_id'], tool_inputs)
        result_dataset_id = run_tool_result['outputs'][0]['id']

        blast_chromosome_mapping = config['blast']['blast_result_to_vcf_chromosome_mapping']

        blast_result_dataset = gi.datasets.download_dataset(result_dataset_id)
        blast_result_json = []

        blast_result_lines = blast_result_dataset.decode('utf-8').split("\n")
        for line in blast_result_lines:
            if line != "":
                line_parts = line.split("\t")
                _chromosome_vcf = str(blast_chromosome_mapping[line_parts[1]])
                _single_blast_hit = {
                    'chromosome': _chromosome_vcf,
                    'percentage_of_identical_matches': line_parts[2],
                    'alignment_length': line_parts[3],
                    'number_of_mismatches': line_parts[4],
                    'number_of_gap_openings': line_parts[5],
                    'start_of_alignment_in_query': line_parts[6],
                    'end_of_alignment_in_query': line_parts[7],
                    'start_of_alignment_in_subject': line_parts[8],
                    'end_of_alignment_in_subject': line_parts[9],
                    'e_value': line_parts[10],
                    'bit_score': line_parts[11],
                    'snp_count': gd.count_variants_in_window(_chromosome_vcf, int(line_parts[8]), int(line_parts[9]))
                }
                blast_result_json.append(_single_blast_hit)
        
        return jsonify({
            'success': True,
            'blast_result': str(blast_result_dataset.decode('utf-8')),
            'blast_hits': blast_result_json
        })


    model_chromosome = api.model('Chromosome', {
        'id': fields.String,
        'label': fields.String,
        'start': fields.Integer,
        'end': fields.Integer,
        'centromere_position': fields.Integer,
        'number_of_variants': fields.Integer,
    })

    model_features = api.model('ConfigurationFeatures', {
        'blast': fields.Boolean,
        'pca': fields.Boolean
    })

    model_configuration = api.model('Configuration', {
        'chromosomes': fields.List(fields.Nested(model_chromosome)),
        'ploidy': fields.Integer,
        'count_variants': fields.Integer,
        'count_genotypes': fields.Integer,
        'count_elements': fields.Integer,
        'samples': fields.List(fields.String),
        'features': fields.Nested(model_features),
        'gff3': fields.Raw(),
        'dataset_descriptions': fields.Raw()
    })

    @api.route('/configuration')
    class Configuration(Resource):
        @api.marshal_with(model_configuration)
        def get(self):
            features = {
                'blast': config['blast']['active'],
                'pca': True,
            }

            samples, _ = gd.map_vcf_sample_ids_to_input_sample_ids(gd.samples.tolist())

            result = {
                'chromosomes': gd.list_of_chromosomes,
                'ploidy': gd.ploidy,
                'count_genotypes': len(gd.samples),
                'count_variants': len(gd.pos),
                'count_elements': len(gd.samples) * len(gd.pos),
                'samples': samples,
                'gff3': ad.metadata_gff3,
                'features': features,
                'dataset_descriptions': dict(config['metadata'])
            }

            #return jsonify(result)
            return result


    '''
    @app.route("/configuration", methods = ['GET', 'POST', 'OPTIONS'])
    def __configuration():

        features = {
            'blast': config['blast']['active'],
            'pca': True,
        }

        samples, _ = gd.map_vcf_sample_ids_to_input_sample_ids(gd.samples.tolist())

        result = {
            'ploidy': gd.ploidy,
            'count_genotypes': len(gd.samples),
            'count_variants': len(gd.pos),
            'count_elements': len(gd.samples) * len(gd.pos),
            'chromosomes': gd.list_of_chromosomes,
            'samples': samples,
            'gff3': ad.metadata_gff3,
            'features': features,
            'dataset_descriptions': dict(config['metadata'])
        }

        return jsonify(result)
    '''


    @app.route("/chromosomes", methods = ['GET', 'POST', 'OPTIONS'])
    def __chromosomes():
        return jsonify(gd.list_of_chromosomes)


    @app.route("/samples", methods = ['GET', 'POST', 'OPTIONS'])
    def __samples():
        return jsonify(gd.samples.tolist())


    @app.route("/genes", methods = ['GET', 'POST', 'OPTIONS'])
    def __genes():
        result = {
            'genes': ad.genes_list,
        }
        return jsonify(result)


    @app.route("/", methods = ['GET', 'POST', 'OPTIONS'])
    def __home():
        return 'Divbrowse server is running'



    @app.errorhandler(ApiError)
    def handle_api_error(error):
        response = jsonify(error.to_dict())
        response.status_code = error.status_code
        return response



    @app.after_request
    def after_request(response):
        header = response.headers
        header['Access-Control-Allow-Origin'] = '*'
        header['Access-Control-Allow-Headers'] = 'Accept, Accept-CH, Accept-Charset, Accept-Datetime, Accept-Encoding, Accept-Ext, Accept-Features, Accept-Language, Accept-Params, Accept-Ranges, Access-Control-Allow-Credentials, Access-Control-Allow-Headers, Access-Control-Allow-Methods, Access-Control-Allow-Origin, Access-Control-Expose-Headers, Access-Control-Max-Age, Access-Control-Request-Headers, Access-Control-Request-Method, Age, Allow, Alternates, Authentication-Info, Authorization, C-Ext, C-Man, C-Opt, C-PEP, C-PEP-Info, CONNECT, Cache-Control, Compliance, Connection, Content-Base, Content-Disposition, Content-Encoding, Content-ID, Content-Language, Content-Length, Content-Location, Content-MD5, Content-Range, Content-Script-Type, Content-Security-Policy, Content-Style-Type, Content-Transfer-Encoding, Content-Type, Content-Version, Cookie, Cost, DAV, DELETE, DNT, DPR, Date, Default-Style, Delta-Base, Depth, Derived-From, Destination, Differential-ID, Digest, ETag, Expect, Expires, Ext, From, GET, GetProfile, HEAD, HTTP-date, Host, IM, If, If-Match, If-Modified-Since, If-None-Match, If-Range, If-Unmodified-Since, Keep-Alive, Label, Last-Event-ID, Last-Modified, Link, Location, Lock-Token, MIME-Version, Man, Max-Forwards, Media-Range, Message-ID, Meter, Negotiate, Non-Compliance, OPTION, OPTIONS, OWS, Opt, Optional, Ordering-Type, Origin, Overwrite, P3P, PEP, PICS-Label, POST, PUT, Pep-Info, Permanent, Position, Pragma, ProfileObject, Protocol, Protocol-Query, Protocol-Request, Proxy-Authenticate, Proxy-Authentication-Info, Proxy-Authorization, Proxy-Features, Proxy-Instruction, Public, RWS, Range, Referer, Refresh, Resolution-Hint, Resolver-Location, Retry-After, Safe, Sec-Websocket-Extensions, Sec-Websocket-Key, Sec-Websocket-Origin, Sec-Websocket-Protocol, Sec-Websocket-Version, Security-Scheme, Server, Set-Cookie, Set-Cookie2, SetProfile, SoapAction, Status, Status-URI, Strict-Transport-Security, SubOK, Subst, Surrogate-Capability, Surrogate-Control, TCN, TE, TRACE, Timeout, Title, Trailer, Transfer-Encoding, UA-Color, UA-Media, UA-Pixels, UA-Resolution, UA-Windowpixels, URI, Upgrade, User-Agent, Variant-Vary, Vary, Version, Via, Viewport-Width, WWW-Authenticate, Want-Digest, Warning, Width, X-Content-Duration, X-Content-Security-Policy, X-Content-Type-Options, X-CustomHeader, X-DNSPrefetch-Control, X-Forwarded-For, X-Forwarded-Port, X-Forwarded-Proto, X-Frame-Options, X-Modified, X-OTHER, X-PING, X-PINGOTHER, X-Powered-By, X-Requested-With'
        return response

    
    return app